#include <iostream>
#include <limits>
#include "tsm.h"
#include "ConvexHull2D.h"
#include "Line2D.h"

using namespace std;
using namespace gmtl;

#define ASSIGN_MAT(M, u0, u3, u6, u1, u4, u7, u2, u5, u8) { \
	M[0][0] = u0; M[0][1] = u3; M[0][2] = u6; \
	M[1][0] = u1; M[1][1] = u4; M[1][2] = u7; \
	M[2][0] = u2; M[2][1] = u5; M[2][2] = u8; \
	}

#define DET2(a, b, c, d) ((a) * (d) - (b) * (c))

#define DOT2(u, v) (u[0] * v[0] + u[1] * v[1])

void intersect(float i[2], float g0[3], float g1[3], float h0[3], float h1[3]);
void map_Trapezoid_To_Square(float TR[3][3], float t0[2], float t1[2], float t2[2], float t3[2]);
void normalizeTrapezoid(Matrix44f& N, Vec2f& t0, Vec2f& t1, Vec2f& t2, Vec2f& t3);
void computeTsmMatrix(Matrix44f& N_T, Camera& eye, Camera& light,
					  float tsmDistance, float percentage, int shadowmapHeight);
void setTrapezoidResolution(Matrix33f& N,
							float percentage, float d1, 
							Line2D& bottom, Line2D& top, 
							Line2D& lineofsight, Vec2f eyeTrans[8]);
float ccw(const Vec2f& p, const Vec2f& q, const Vec2f& r); //define in Convexhull2D.cpp

void normalizeTrapezoid(Matrix33f& N_T, Vec2f& t0, Vec2f& t1, Vec2f& t2, Vec2f& t3) {
	float TR[3][3];

	map_Trapezoid_To_Square(TR, t0.mData, t1.mData, t2.mData, t3.mData);

	N_T.mData[0] = TR[0][0]; N_T.mData[3] = TR[0][1]; N_T.mData[6] = TR[0][2];
	N_T.mData[1] = TR[1][0]; N_T.mData[4] = TR[1][1]; N_T.mData[7] = TR[1][2];
	N_T.mData[2] = TR[2][0]; N_T.mData[5] = TR[2][1]; N_T.mData[8] = TR[2][2];
}

void computeTsmMatrix(Matrix44f& N_T, Camera& eye, Camera& light,
					  float tsmDistance, float percentage, int shadowmapHeight) {
	Camera visEye = eye;

	visEye.setFarPlaneDistance(tsmDistance);

	Vec2f eyeTrans[8];
	Vec2f eyeVisTrans[8];

	Matrix44f Ml = light.m_P * light.m_C;

	Vec4f* p1 = eye.m_vtVerticesTransformed;	
	Vec4f* p2 = visEye.m_vtVerticesTransformed;

	for(int i = 0; i < 8; i++) {
		Vec4f v = Ml * p1[i];
		eyeTrans[i][0] = v[0] / v[3];
		eyeTrans[i][1] = v[1] / v[3];

		v = Ml * p2[i];
		eyeVisTrans[i][0] = v[0] / v[3];
		eyeVisTrans[i][1] = v[1] / v[3];
	}
	
	Vec4f u = 0.5f * (p1[2] - p1[0]) + p1[0];
	Vec4f v = 0.5f * (p1[6] - p1[4]) + p1[4];
	u = Ml * u;
	v = Ml * v;		
	
	Line2D losEye(Vec2f(u[0] / u[3], u[1] / u[3]), Vec2f(v[0] / v[3], v[1] / v[3]));

	Convexhull2D ch1(eyeTrans, 8);
	Convexhull2D ch2(eyeVisTrans, 8);

	Line2D orthoLosEye = losEye.ortho();	
	Vec2f maxPointA = orthoLosEye.maxPointToLine(ch1.vertices);

	Line2D bottom = orthoLosEye.parallel(maxPointA);
	Vec2f maxPointB = bottom.maxPointToLine(ch1.vertices);
	
	Line2D top = bottom.parallel(maxPointB);	
	Vec2f maxPointC = top.maxPointToLine(ch2.vertices);

	float d1 = fabs(top.distpoint(maxPointC));

	float maxArea = 0.0f;
	float f;

	float fInc = 100.0f / shadowmapHeight;

	float g = percentage;

	Matrix33f N;

	for(g = percentage; g >= 0.0f; g -= fInc) {
		setTrapezoidResolution(N, g, d1, bottom, top, losEye, eyeTrans);

		vector<Vec2f> ch(ch2.size());
		
		for(int i = 0; i < ch2.size(); i++) {
			Vec3f v = N * Vec3f(ch2.vertices[i][0], ch2.vertices[i][1], 1.0f);
			ch[i] = Vec2f(v[0] / v[2], v[1] / v[2]);
		}

		f = area(ch);

		if(f < maxArea) {
			g += fInc;
			break;
		} else {
			maxArea = f;
		}
	}

	N_T.mData[0] = N.mData[0]; N_T.mData[4] = N.mData[3]; N_T.mData[ 8] =  0.0f; N_T.mData[12] = N.mData[6];
	N_T.mData[1] = N.mData[1]; N_T.mData[5] = N.mData[4]; N_T.mData[ 9] =  0.0f; N_T.mData[13] = N.mData[7];
	N_T.mData[2] =       0.0f; N_T.mData[6] =       0.0f; N_T.mData[10] =  1.0f; N_T.mData[14] =       0.0f;
	N_T.mData[3] = N.mData[2]; N_T.mData[7] = N.mData[5]; N_T.mData[11] =  0.0f; N_T.mData[15] = N.mData[8];	
}
void setTrapezoidResolution(Matrix33f& N,
							float percentage, float d1, 
							Line2D& bottom, Line2D& top, 
							Line2D& lineofsight, Vec2f eyeTrans[8]) {
	
	Vec2f s = bottom.getA();
	float l = fabs(top.distpoint(s));
	float d2 = -((2.0f * percentage) / 100.0f - 1.0f);
	float n = (l * d1 + l * d1 * d2) / (l - 2.0f * d1 - l * d2);

	Vec2f u, v, w, i;
	intersect(u, top, lineofsight);
	intersect(v, bottom, lineofsight);
	w = v - u;

	float t = -n / gmtl::length(w);
	i = u + w * t;

	float min1, min2, f;
	int i0 = 0, i1 = 0;
	min1 = min2 = numeric_limits<float>::max();
	w = i - u;
	
	gmtl::normalize(w);
		
	for(int j = 4; j < 8; j++) {
		Vec2f x = eyeTrans[j] - i;
		
		gmtl::normalize(x);
		
		f = static_cast<float>(acos(gmtl::dot(x, w)));		
		if(ccw(u, v, eyeTrans[j]) < 0.0f) {
			if(f < min1) {
				min1 = f;
				i0 = j;
			}
		} else {
			if(f < min2) {
				min2 = f;
				i1 = j;
			}
		}
	}

	Vec2f left = eyeTrans[i0], right = eyeTrans[i1];

	Line2D h(i, right);
	Line2D g(i, left);

	Vec2f t0, t1, t2, t3;
	if(!intersect(t0, bottom, h)) cerr << "couldn't calc t0" << endl;
	if(!intersect(t1, bottom, g)) cerr << "couldn't calc t1" << endl;
	if(!intersect(t2, top, g)) cerr << "couldn't calc t2" << endl;
	if(!intersect(t3, top, h)) cerr << "couldn't calc t3" << endl;

	normalizeTrapezoid(N, t0, t1, t2, t3);
}



void intersect(float i[2], float g0[3], float g1[3], float h0[3], float h1[3]) {
	float a, b;

	i[0] = i[1] = 1.0f / DET2(g0[0] - g1[0], g0[1] - g1[1], h0[0] - h1[0], h0[1] - h1[1]);

	a = DET2(g0[0], g0[1], g1[0], g1[1]);
	b = DET2(h0[0], h0[1], h1[0], h1[1]);

	i[0] *=	DET2(a, g0[0] - g1[0], b, h0[0] - h1[0]);
	i[1] *=	DET2(a, g0[1] - g1[1], b, h0[1] - h1[1]);
}

void map_Trapezoid_To_Square(float TR[3][3], float t0[2], float t1[2], float t2[2], float t3[2]) {
	float i[2], a, b, c, d;

	//M1 = R * T1
	a = 0.5f * (t2[0] - t3[0]);
	b = 0.5f * (t2[1] - t3[1]);

	ASSIGN_MAT(TR, a  ,  b  , a * a + b * b,
	               b  , -a  , a * b - b * a,
	              0.0f, 0.0f, 1.0f);

	//M2 = T2 * M1 = T2 * R * T1
	intersect(i, t0, t3, t1, t2);

	TR[0][2] = -DOT2(TR[0], i);
	TR[1][2] = -DOT2(TR[1], i);

	//M1 = H * M2 = H * T2 * R * T1
	a = DOT2(TR[0], t2) + TR[0][2];
	b = DOT2(TR[1], t2) + TR[1][2];
	c = DOT2(TR[0], t3) + TR[0][2];
	d = DOT2(TR[1], t3) + TR[1][2];

	a = -(a + c) / (b + d);

	TR[0][0] += TR[1][0] * a;
	TR[0][1] += TR[1][1] * a;
	TR[0][2] += TR[1][2] * a;

	//M2 = S1 * M1 = S1 * H * T2 * R * T1
	a = 1.0f / (DOT2(TR[0], t2) + TR[0][2]);
	b = 1.0f / (DOT2(TR[1], t2) + TR[1][2]);

	TR[0][0] *= a; TR[0][1] *= a; TR[0][2] *= a;
	TR[1][0] *= b; TR[1][1] *= b; TR[1][2] *= b;

	//M1 = N * M2 = N * S1 * H * T2 * R * T1
	TR[2][0] = TR[1][0]; TR[2][1] = TR[1][1]; TR[2][2] = TR[1][2];
	TR[1][2] += 1.0f;

	//M2 = T3 * M1 = T3 * N * S1 * H * T2 * R * T1
	a = DOT2(TR[1], t0) + TR[1][2];
	b = DOT2(TR[2], t0) + TR[2][2];
	c = DOT2(TR[1], t2) + TR[1][2];
	d = DOT2(TR[2], t2) + TR[2][2];

	a = -0.5f * (a / b + c / d);

	TR[1][0] += TR[2][0] * a;
	TR[1][1] += TR[2][1] * a;
	TR[1][2] += TR[2][2] * a;

	//M1 = S2 * M2 = S2 * T3 * N * S1 * H * T2 * R * T1
	a = DOT2(TR[1], t0) + TR[1][2];
	b = DOT2(TR[2], t0) + TR[2][2];

	c = -b / a;

	TR[1][0] *= c; TR[1][1] *= c; TR[1][2] *= c;
}